use crate::route::ToAPI;
use serde::{Deserialize, Serialize};
use serde_json::json;
use sqlx::FromRow;

use crate::get_pg;

use super::track::Track;

#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct Artist {
    pub id: uuid::Uuid,
    pub name: String,
}

impl Artist {
    pub async fn create(name: &str) -> Self {
        sqlx::query_as("INSERT INTO artist (name) VALUES ($1) RETURNING id, name")
            .bind(name)
            .fetch_one(get_pg!())
            .await
            .unwrap()
    }

    pub async fn get(id: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM artist WHERE id = $1")
            .bind(id)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn find(name: &str) -> Option<Self> {
        sqlx::query_as("SELECT * FROM artist WHERE name = $1")
            .bind(name)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn find_all() -> Vec<Self> {
        sqlx::query_as("SELECT * FROM artist")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn find_regex_col(col: &str, query: &str) -> Vec<Self> {
        sqlx::query_as(&format!("SELECT * FROM artist WHERE {col} ~* $1"))
            .bind(query)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn remove(&self) {
        sqlx::query("DELETE FROM artist WHERE id = $1")
            .bind(self.id)
            .fetch_one(get_pg!())
            .await
            .unwrap();
    }

    /// Gets the image of an artist or `None` if it can't be found.
    ///
    /// This function gets a track from the artist. It then expects the folder structure to be `Artist/Album/Track.ext` and searches for an image file named `artist` in the artist folder.
    pub async fn get_image_of(id: &uuid::Uuid) -> Option<String> {
        let track_path = Track::find_first_of_artist(id).await?.path;
        let track_path = std::path::Path::new(&track_path);

        let artist_path = track_path.parent()?.parent()?;

        for ext in ["png", "jpg", "jpeg", "avif", "webp"] {
            let cover_file = artist_path.join(format!("artist.{ext}"));

            if cover_file.exists() {
                return Some(cover_file.to_str().unwrap().to_string());
            }
        }

        None
    }
}

impl ToAPI for Artist {
    async fn api(&self) -> serde_json::Value {
        json!({
            "id": &self.id,
            "name": &self.name,
            "image": if Artist::get_image_of(&self.id).await.is_some() {
                Some(format!("/artist/{}/image", self.id))
            } else {
                None
            }
        })
    }
}
