use std::{
    path::{Path, PathBuf},
    str::FromStr,
};

use album::Album;
use artist::Artist;
use serde_json::json;
use track::Track;
use walkdir::WalkDir;

use crate::{cache::RouteCache, get_pg};

pub mod album;
pub mod artist;
pub mod event;
pub mod metadata;
pub mod playlist;
pub mod search;
pub mod track;
pub mod user;

/// Checks if a file has a music file extension
fn is_music_file(path: &Path) -> bool {
    if let Some(extension) = path.extension() {
        match extension.to_str().unwrap_or("").to_lowercase().as_str() {
            "flac" | "wav" | "aac" | "ogg" | "opus" => return true,
            _ => return false,
        }
    }
    false
}

pub struct Libary {
    root_dir: PathBuf,
}

impl Libary {
    pub const fn new(root_dir: PathBuf) -> Self {
        Self { root_dir }
    }

    pub async fn find_or_create_artist(&self, artist: &str) -> uuid::Uuid {
        if let Some(artist) = Artist::find(artist).await {
            artist.id
        } else {
            Artist::create(artist).await.id
        }
    }

    pub async fn find_or_create_album(
        &self,
        album: &str,
        artist_id: Option<uuid::Uuid>,
    ) -> uuid::Uuid {
        if let Some(album) = Album::find(album, artist_id).await {
            album.id
        } else {
            Album::create(album, artist_id).await.id
        }
    }

    pub async fn add_path_to_library(&self, path: &str) {
        // search for path already present
        if Track::of_path(path).await.is_some() {
            return;
        }

        log::info!("Adding {path} to library");

        // add track to library
        let metadata = metadata::get_metadata(path);

        let mut entry = json!({
            "path": path,
        });

        if let Some(meta) = &metadata {
            if let Some(artist) = meta.artist() {
                let artist_id = self.find_or_create_artist(artist).await;
                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("artist".into(), artist_id.to_string().into());
            } else {
                log::warn!("{path} has no artist");
            }

            if let Some(album) = meta.album() {
                let album_id = self
                    .find_or_create_album(
                        album,
                        entry
                            .as_object()
                            .unwrap()
                            .get("artist")
                            .map(|x| uuid::Uuid::from_str(x.as_str().unwrap()).unwrap()),
                    )
                    .await;

                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("album".into(), album_id.to_string().into());
            } else {
                log::warn!("{path} has no album and will be treated as single");
            }

            if let Some(title) = meta.title() {
                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("title".into(), title.into());
            }
        }

        if let Some(metadata) = metadata {
            entry
                .as_object_mut()
                .unwrap()
                .insert("meta".into(), metadata.0);
        }

        // if no title in metadata use file name
        if entry.as_object().unwrap().get("title").is_none() {
            entry.as_object_mut().unwrap().insert(
                "title".into(),
                std::path::Path::new(path)
                    .file_stem()
                    .unwrap()
                    .to_str()
                    .unwrap()
                    .into(),
            );
        }

        Track::create(entry.as_object().unwrap()).await;
    }

    pub async fn get_artists(&self) -> Vec<Artist> {
        sqlx::query_as("SELECT * FROM artist")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn get_albums_by_artist(&self, artist: &uuid::Uuid) -> Vec<Album> {
        sqlx::query_as("SELECT * FROM album WHERE artist = $1")
            .bind(artist)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn get_singles_by_artist(&self, artist: &uuid::Uuid) -> Vec<Track> {
        sqlx::query_as("SELECT * FROM track WHERE album IS NULL AND artist = $1")
            .bind(artist)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn get_album_by_id(&self, album: &uuid::Uuid) -> Option<Album> {
        Album::get(album).await
    }

    pub async fn get_track_by_id(&self, track_id: &uuid::Uuid) -> Option<Track> {
        Track::get(track_id).await
    }

    pub async fn reload_metadata(&self, track_id: &uuid::Uuid) -> Result<(), ()> {
        let track = Track::get(track_id).await.ok_or_else(|| ())?;
        let path = &track.path;
        log::info!("Rescanning metadata for {path}");

        let metadata = metadata::get_metadata(path);

        let mut update = json!({});

        if let Some(meta) = &metadata {
            if let Some(artist) = meta.artist() {
                let artist_id = self.find_or_create_artist(artist).await;
                update
                    .as_object_mut()
                    .unwrap()
                    .insert("artist".into(), artist_id.to_string().into());
            } else {
                log::warn!("{path} has no artist");
            }

            if let Some(album) = meta.album() {
                let album_id = self
                    .find_or_create_album(
                        album,
                        update
                            .as_object()
                            .unwrap()
                            .get("artist")
                            .map(|x| uuid::Uuid::from_str(x.as_str().unwrap()).unwrap()),
                    )
                    .await;
                update
                    .as_object_mut()
                    .unwrap()
                    .insert("album".into(), album_id.to_string().into());
            } else {
                log::warn!("{path} has no album and will be treated as single");
            }

            if let Some(title) = meta.title() {
                update
                    .as_object_mut()
                    .unwrap()
                    .insert("title".into(), title.into());
            }
        }

        if let Some(metadata) = metadata {
            update
                .as_object_mut()
                .unwrap()
                .insert("meta".into(), metadata.0);
        }

        // if no title in metadata use file name
        if update.as_object().unwrap().get("title").is_none() {
            update.as_object_mut().unwrap().insert(
                "title".into(),
                std::path::Path::new(&path)
                    .file_stem()
                    .unwrap()
                    .to_str()
                    .unwrap()
                    .into(),
            );
        }

        track.update(&update).await.unwrap();

        Ok(())
    }

    pub async fn clean_lost_files(&self) {
        // tracks
        for track in Track::find_all().await {
            if !std::path::Path::new(&track.path).exists() {
                log::info!("Cleaning lost {}", track.path);
                track.remove().await;
            }
        }
        // albums
        for album in Album::find_all().await {
            if Track::find_of_album(&album.id).await.is_empty() {
                log::info!("Cleaning album {} with no tracks", album.title);
                album.remove().await;
            }
        }
        // artists
        for artist in Artist::find_all().await {
            if Track::find_first_of_artist(&artist.id).await.is_none()
                && Album::find_of_artist(&artist.id).await.is_empty()
            {
                log::info!("Cleaning artist {} with no tracks or albums", artist.name);
                artist.remove().await;
            }
        }
    }

    pub async fn rescan(&self, cache: &RouteCache) {
        cache.invalidate("albums", "latest").await;

        log::info!("Rescanning library");
        for entry in WalkDir::new(self.root_dir.clone())
            .follow_links(true)
            .into_iter()
            .filter_map(std::result::Result::ok)
        {
            let path = entry.path();
            if path.is_file() && is_music_file(path) {
                let path = path.to_string_lossy().to_string();
                self.add_path_to_library(&path).await;
            }
        }
    }
}
