use serde::{Deserialize, Serialize};
use sqlx::FromRow;

use crate::{get_pg, library::user::User};

#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct Playlist {
    pub id: String,
    pub owner: String,
    pub title: String,
    pub visibility: Visibility,
    pub tracks: Vec<uuid::Uuid>,
}

#[derive(Debug, Clone, Serialize, Deserialize, sqlx::Type)]
#[sqlx(type_name = "visibility", rename_all = "lowercase")]
pub enum Visibility {
    Private,
    Public,
}

impl Playlist {
    pub async fn create(
        owner: User,
        title: &str,
        visibility: Visibility,
        tracks: &[uuid::Uuid],
    ) -> Option<Self> {
        sqlx::query_as("INSERT INTO playlist (owner, title, visibility, tracks) VALUES ($1, $2, $3, $4) RETURNING *")
        .bind(owner.username)
        .bind(title)
        .bind(visibility)
        .bind(tracks)
        .fetch_one(get_pg!()).await.ok()
    }

    pub async fn find_regex_col(col: &str, query: &str) -> Vec<Self> {
        sqlx::query_as(&format!("SELECT * FROM playlist WHERE {col} ~* $1"))
            .bind(query)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn get(id: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM playlist WHERE id = $1")
            .bind(id)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn of(u: &User) -> Vec<Self> {
        sqlx::query_as("SELECT * FROM playlist WHERE owner = $1")
            .bind(&u.username)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }
}

impl Playlist {
    pub async fn api(&self) -> serde_json::Value {
        serde_json::json!({
            "id": self.id,
            "owner": self.owner,
            "visibility": serde_json::to_value(&self.visibility).unwrap(),
            "title": self.title,
            "tracks": self.tracks
        })
    }
}
