use serde::{Deserialize, Serialize};
use serde_json::json;
use sqlx::prelude::FromRow;
use std::{collections::HashSet, str::FromStr};

use crate::{
    get_pg,
    library::album::Album,
    route::{to_uuid, ToAPI},
};

use super::{event::Event, metadata::AudioMetadata, user::User};

#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct Track {
    pub id: uuid::Uuid,
    pub path: String,
    pub title: String,
    pub date_added: chrono::DateTime<chrono::Utc>,
    pub album: Option<uuid::Uuid>,
    pub artist: Option<uuid::Uuid>,
    pub meta: Option<serde_json::Value>,
}

impl Track {
    pub async fn create(data: &serde_json::Map<String, serde_json::Value>) {
        sqlx::query(
            "INSERT INTO track (path, title, meta, album, artist) VALUES ($1, $2, $3, $4, $5)",
        )
        .bind(data.get("path").unwrap().as_str().unwrap().to_string())
        .bind(data.get("title").unwrap().as_str().unwrap().to_string())
        .bind(data.get("meta"))
        .bind(
            data.get("album")
                .map(|x| to_uuid(x.as_str().unwrap()).unwrap()),
        )
        .bind(
            data.get("artist")
                .map(|x| to_uuid(x.as_str().unwrap()).unwrap()),
        )
        .execute(get_pg!())
        .await
        .unwrap();
    }

    pub async fn of_path(path: &str) -> Option<Self> {
        sqlx::query_as("SELECT * FROM track WHERE path = $1")
            .bind(path)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn find_recently_added() -> Vec<Self> {
        sqlx::query_as("SELECT * FROM track ORDER BY date_added DESC LIMIT 90")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn get(id: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM track WHERE id = $1")
            .bind(id)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn update(&self, update_set: &serde_json::Value) -> Option<()> {
        let map = update_set.as_object()?;
        let artist = map
            .get("artist")
            .map(|x| uuid::Uuid::from_str(x.as_str().unwrap()).unwrap());
        let album = map
            .get("album")
            .map(|x| uuid::Uuid::from_str(x.as_str().unwrap()).unwrap());
        let title = map.get("title")?.as_str()?;
        let meta = map.get("meta");

        sqlx::query(
            "UPDATE track SET artist = $1, album = $2, title = $3, meta = $4 WHERE id = $5;
",
        )
        .bind(artist)
        .bind(album)
        .bind(title)
        .bind(meta)
        .bind(self.id)
        .fetch_one(get_pg!())
        .await
        .unwrap();

        Some(())
    }

    pub async fn find_all() -> Vec<Self> {
        sqlx::query_as("SELECT * FROM track")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn remove(&self) {
        sqlx::query("DELETE FROM track WHERE id = $1")
            .bind(self.id)
            .fetch_one(get_pg!())
            .await
            .unwrap();
    }

    pub async fn find_of_album(album: &uuid::Uuid) -> Vec<Self> {
        sqlx::query_as("SELECT * FROM track WHERE album = $1")
            .bind(album)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn get_latest_of_user(u: &User) -> Vec<Self> {
        let latest_events = Event::get_latest_events_of(u).await;
        let mut ids = HashSet::new();

        for event in latest_events {
            if !ids.contains(&event.track) {
                ids.insert(event.track.clone());
            }
        }

        sqlx::query_as("SELECT * FROM track WHERE id = ANY($1)")
            .bind(ids.into_iter().collect::<Vec<_>>())
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Transcode audio to OPUS with `bitrate`
    pub fn get_opus(&self, bitrate: u32) -> Option<String> {
        self.transcode("libopus", bitrate, "opus")
    }

    /// Transcode audio to AAC with `bitrate`
    pub fn get_aac(&self, bitrate: u32) -> Option<String> {
        self.transcode("aac", bitrate, "aac")
    }

    /// Transcode audio
    pub fn transcode(&self, codec: &str, bitrate: u32, ext: &str) -> Option<String> {
        let transcoded = format!("./data/transcode/{codec}/{bitrate}/{}.{ext}", self.id);

        if std::path::Path::new(&transcoded).exists() {
            return Some(transcoded);
        }

        log::info!("Transcoding {} to {} {}", self.id, codec, bitrate);

        std::fs::create_dir_all(format!("./data/transcode/{codec}/{bitrate}")).unwrap();

        let out = std::process::Command::new("ffmpeg")
            .arg("-i")
            .arg(&self.path)
            .arg("-c:a")
            .arg(codec)
            .arg("-b:a")
            .arg(format!("{bitrate}k"))
            .arg(&transcoded)
            .output()
            .unwrap();

        if !out.status.success() {
            return None;
        }

        Some(transcoded)
    }

    pub async fn find_regex_col(col: &str, query: &str) -> Vec<Self> {
        sqlx::query_as(&format!("SELECT * FROM track WHERE {col} ~* $1"))
            .bind(query)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Find tracks with no album or artist
    pub async fn get_orphans() -> Vec<Track> {
        sqlx::query_as("SELECT * FROM track WHERE artist IS NULL AND album IS NULL")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn find_first_of_artist(artist: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM track WHERE artist = $1 LIMIT 1")
            .bind(artist)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn find_first_of_album(album: &uuid::Uuid) -> Option<Self> {
        sqlx::query_as("SELECT * FROM track WHERE album = $1 LIMIT 1")
            .bind(album)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }
}

impl ToAPI for Track {
    async fn api(&self) -> serde_json::Value {
        // todo : fix
        let (cover, album_title, album_id) = if let Some(album_ref) = &self.album {
            let album = Album::get(album_ref).await.unwrap();

            (album.get_cover().await.is_some(), album.title, album.id)
        } else {
            (false, String::new(), uuid::Uuid::nil())
        };

        let artist_title = if let Some(artist_ref) = &self.artist {
            let res: (String,) = sqlx::query_as("SELECT name FROM artist WHERE id = $1")
                .bind(artist_ref)
                .fetch_one(get_pg!())
                .await
                .unwrap();
            Some(res.0)
        } else {
            None
        };

        json!({
            "id": self.id,
            "title": self.title,
            "track_number": self.meta.as_ref().map(|x| AudioMetadata(x.clone()).track_number()),
            "meta": serde_json::to_value(&self.meta).unwrap(),
            "album_id": self.album,
            "album": album_title,
            "cover": if cover {
                Some(format!("/album/{album_id}/cover"))
            } else {
                None
            },
            "artist_id": self.artist,
            "artist": artist_title
        })
    }
}
