use crate::route::ToAPI;
use data_encoding::HEXUPPER;
use rand::RngCore;
use serde::{Deserialize, Serialize};
use serde_json::json;
use sqlx::FromRow;

use crate::get_pg;

fn gen_token(token_length: usize) -> String {
    let mut token_bytes = vec![0u8; token_length];

    rand::thread_rng().fill_bytes(&mut token_bytes);

    HEXUPPER.encode(&token_bytes)
}

#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct User {
    pub username: String,
    pub password: String,
    pub user_role: UserRole,
}

#[derive(Debug, Clone, Serialize, Deserialize, sqlx::Type)]
#[sqlx(type_name = "user_role", rename_all = "lowercase")]
pub enum UserRole {
    Regular,
    Admin,
}

impl User {
    pub async fn find(username: &str) -> Option<Self> {
        sqlx::query_as("SELECT * FROM users WHERE username = $1")
            .bind(username)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn create(username: &str, password: &str, role: UserRole) -> Option<Self> {
        if Self::find(username).await.is_some() {
            return None;
        }

        let u = Self {
            username: username.to_string(),
            password: bcrypt::hash(password, bcrypt::DEFAULT_COST).unwrap(),
            user_role: role,
        };

        sqlx::query("INSERT INTO users (username, \"password\", user_role) VALUES ($1, $2, $3)")
            .bind(&u.username)
            .bind(&u.password)
            .bind(&u.user_role)
            .execute(get_pg!())
            .await
            .unwrap();

        Some(u)
    }

    pub async fn login(username: &str, password: &str) -> Option<(Session, UserRole)> {
        let u = Self::find(username).await?;

        if !u.verify_pw(password) {
            return None;
        }

        Some((u.session().await, u.user_role))
    }

    /// Change the password of a `User`
    pub async fn passwd(self, old: &str, new: &str) -> Result<(), ()> {
        if self.verify_pw(old) {
            sqlx::query("UPDATE users SET \"password\" = $1 WHERE username = $2;")
                .bind(bcrypt::hash(new, bcrypt::DEFAULT_COST).unwrap())
                .bind(&self.username)
                .fetch_one(get_pg!())
                .await
                .unwrap();

            return Ok(());
        }

        Err(())
    }

    pub async fn find_all() -> Vec<Self> {
        sqlx::query_as("SELECT * FROM users")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    pub async fn session(&self) -> Session {
        sqlx::query_as(
            "INSERT INTO user_session (token, \"user\") VALUES ($1, $2) RETURNING id, token, \"user\"",
        )
        .bind(gen_token(64))
        .bind(&self.username)
        .fetch_one(get_pg!())
        .await
        .unwrap()
    }

    pub const fn is_admin(&self) -> bool {
        matches!(self.user_role, UserRole::Admin)
    }

    pub fn verify_pw(&self, password: &str) -> bool {
        bcrypt::verify(password, &self.password).unwrap()
    }
}

impl ToAPI for User {
    async fn api(&self) -> serde_json::Value {
        json!({
            "username": self.username,
            "role": self.user_role
        })
    }
}

#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct Session {
    pub id: uuid::Uuid,
    pub token: String,
    pub user: String,
}
